import sys

from girder_worker_utils.tee import Tee, tee_stdout, TeeStdErr, TeeStdOut


@tee_stdout
class TeeCapture(Tee):
    def __init__(self, *args, **kwargs):
        self.buf = ''
        super().__init__(*args, **kwargs)

    def write(self, message, **kwargs):
        self.buf += message
        super().write(message, **kwargs)


def test_tee_sys_write_stdout(capfd):
    with TeeStdOut():
        sys.stdout.write('Test String')
        sys.stdout.flush()

    out, err = capfd.readouterr()
    assert out == 'Test String'


def test_tee_print_stdout(capfd):
    with TeeStdOut():
        print('Test String')

    out, err = capfd.readouterr()
    assert out == 'Test String\n'


def test_tee_stdout_sys_write_pass_through_false(capfd):
    with TeeStdOut(pass_through=False):
        sys.stdout.write('Test String')
        sys.stdout.flush()

    out, err = capfd.readouterr()
    assert out == ''


def test_tee_stdout_print_pass_through_false(capfd):
    with TeeStdOut(pass_through=False):
        print('Test String')

    out, err = capfd.readouterr()
    assert out == ''


def test_tee_sys_write_stderr(capfd):
    with TeeStdErr():
        sys.stderr.write('Test String')
        sys.stderr.flush()

    out, err = capfd.readouterr()
    assert err == 'Test String'


def test_tee_stderr_sys_write_pass_through_false(capfd):
    with TeeStdErr(pass_through=False):
        sys.stderr.write('Test String')
        sys.stderr.flush()

    out, err = capfd.readouterr()
    assert err == ''


def test_tee_overwrites_write(capfd):
    with TeeCapture() as o:
        print('Test String')
        assert o.buf == 'Test String\n'

    out, err = capfd.readouterr()
    assert out == 'Test String\n'


def test_tee_overwrites_write_pass_through_false(capfd):
    with TeeCapture(pass_through=False) as o:
        print('Test String')
        assert o.buf == 'Test String\n'

    out, err = capfd.readouterr()
    assert out == ''


def test_tee_reset_function_resets(capfd):
    original_stdout = sys.stdout

    o = TeeCapture()

    print('Test String')
    out, err = capfd.readouterr()

    assert o.buf == 'Test String\n'
    assert out == 'Test String\n'

    o.reset()

    assert sys.stdout == original_stdout


def test_tee_multiple_tee_objects_downstream():
    original_stdout = sys.stdout
    o1, o2, o3 = TeeCapture(), TeeCapture(), TeeCapture()

    assert sys.stdout == o3
    assert sys.stdout._downstream == o2
    assert sys.stdout._downstream._downstream == o1
    assert sys.stdout._downstream._downstream._downstream == original_stdout


def test_tee_multiple_tee_objects_each_recieves_write(capfd):
    original_stdout = sys.stdout
    o1, o2, o3 = TeeCapture(), TeeCapture(), TeeCapture()

    assert sys.stdout == o3
    assert o3._downstream == o2
    assert o2._downstream == o1
    assert o1._downstream == original_stdout

    print('Test String')

    assert o3.buf == 'Test String\n'
    assert o2.buf == 'Test String\n'
    assert o1.buf == 'Test String\n'

    out, err = capfd.readouterr()
    assert out == 'Test String\n'


def test_tee_multiple_tee_objects_reset_o1():
    original_stdout = sys.stdout
    o1, o2, o3 = TeeCapture(), TeeCapture(), TeeCapture()

    o1.reset()

    assert o3._downstream == o2
    assert o2._downstream == original_stdout


def test_tee_multiple_tee_objects_reset_o2():
    original_stdout = sys.stdout
    o1, o2, o3 = TeeCapture(), TeeCapture(), TeeCapture()

    o2.reset()

    assert o3._downstream == o1
    assert o1._downstream == original_stdout


def test_tee_multiple_tee_objects_reset_o3():
    original_stdout = sys.stdout
    o1, o2, o3 = TeeCapture(), TeeCapture(), TeeCapture()

    o3.reset()

    assert sys.stdout == o2
    assert o2._downstream == o1
    assert o1._downstream == original_stdout


def test_tee_multiple_tee_objects_reset_o3_then_o1():
    original_stdout = sys.stdout
    o1, o2, o3 = TeeCapture(), TeeCapture(), TeeCapture()

    o3.reset()

    assert sys.stdout == o2
    assert o2._downstream == o1
    assert o1._downstream == original_stdout

    o1.reset()

    assert sys.stdout == o2
    assert o2._downstream == original_stdout


def test_tee_multiple_tee_objects_reset_o3_then_o1_then_o2():
    original_stdout = sys.stdout
    o1, o2, o3 = TeeCapture(), TeeCapture(), TeeCapture()

    o3.reset()

    assert sys.stdout == o2
    assert o2._downstream == o1
    assert o1._downstream == original_stdout

    o1.reset()

    assert sys.stdout == o2
    assert o2._downstream == original_stdout

    o2.reset()

    assert sys.stdout == original_stdout



from .number import Number


class Integer(Number):
    """Define an integer task parameter.

    >>> @argument('iterations', types.Integer)
    ... def func(iterations=3):
    ...     pass
    """

    paramType = 'integer'

    def __init__(self, *args, **kwargs):
        kwargs['step'] = 1
        super().__init__(*args, **kwargs)

    def serialize(self, value):
        value = super().serialize(value)
        return int(value)



import importlib.metadata

from . import log_utils

try:
    __version__ = importlib.metadata.version(__name__)
except importlib.metadata.PackageNotFoundError:
    # package is not installed
    pass


__license__ = 'Apache 2.0'


# Create and configure our logger
logger = log_utils.setupLogger()


class GirderWorkerPluginABC:
    """
    Abstract base class for Girder Worker plugins. Plugins must descend from this
    class; see the :ref:`plugins` section for more information.
    """

    def __init__(*args, **kwargs):
        pass

    def task_imports(self):
        """Plugins should override this method if they have tasks."""
        return []


class GirderWorkerPlugin(GirderWorkerPluginABC):

    def __init__(self, app, *args, **kwargs):
        self.app = app

    def task_imports(self):
        return ['girder_worker.tasks']



from copy import deepcopy


class Base:
    """Define an abstract type class.

    This class defines the interface expected by the functions in
    :py:mod:`girder_worker.describe`.  All autodescribing parameter
    types should derive from this class.
    """

    #: This is used as a base description object for all types.
    description = {}

    def __init__(self, name, **kwargs):
        """Construct a task description.

        Subclasses can define additional keyword arguments.

        :param str name: The argument name in the task function.
        :param str id: An optional id (defaults to the argument name)
        :param str help: A help string (defaults to the function docstring)
        """
        self.id = kwargs.get('id', name)
        self.name = name
        self.help = kwargs.get('help')

    def set_parameter(self, parameter, **kwargs):
        """Store a parameter instance from a function signature.

        This method is called internally by description decorators.

        :param parameter: The parameter signature from the task
        :type parameter: inspect.Parameter
        """
        self.parameter = parameter

    def has_default(self):
        """Return true if the parameter has a default value."""
        return self.parameter.default is not self.parameter.empty

    def describe(self):
        """Return a type description serialization."""
        copy = deepcopy(self.description)
        copy.setdefault('id', self.id)
        copy.setdefault('name', self.name)
        if self.has_default():
            copy.setdefault('default', {
                'data': self.serialize(self.parameter.default)
            })
        if self.help is not None:
            copy['description'] = self.help
        return copy

    def serialize(self, value):
        """Serialize a python value into a format expected by item_tasks."""
        return value

    def deserialize(self, value):
        """Deserialize a python value from a format provided by item_tasks."""
        return value

    def validate(self, value):
        """Validate a parameter value.

        :raises Exception: if the value is not valid
        """



from .base import Base


class Color(Base):
    """Define a task parameter expecting a color value.

    >>> @argument('background', types.Color)
    ... def func(background):
    ...     pass
    """

    description = {
        'type': 'color',
        'description': 'Provide a color value'
    }

    # TODO: handle normalization and validation



import shutil
import tempfile

from ..transform import Transform


class TemporaryDirectory(Transform):
    """
    This transform returns a temporary directory that is removed on cleanup.
    """

    def _repr_model_(self):
        return self.__class__.__name__

    def transform(self):
        self.temp_dir_path = tempfile.mkdtemp()
        return self.temp_dir_path

    def cleanup(self):
        if hasattr(self, 'temp_dir_path'):
            shutil.rmtree(self.temp_dir_path, ignore_errors=True)



from .base import Base


class Vector(Base):
    """Define a base class that accepts an iterable object."""

    #: The explicit type provided by the subclass
    paramType = None

    #: The class of the elements of the vector
    elementClass = None

    #: A list separator for serialization
    separator = ','

    def __init__(self, *args, **kwargs):
        if self.paramType is None:  # pragma: nocover
            raise NotImplementedError('Subclasses should define paramType')

        if self.elementClass is None:  # pragma: nocover
            raise NotImplementedError('Subclasses should define elementClass')

        self.element = self.elementClass(*args, **kwargs)
        super().__init__(*args, **kwargs)

    def describe(self, *args, **kwargs):
        desc = super().describe(*args, **kwargs)
        desc['type'] = self.paramType
        desc['description'] = 'Provide a list of values'
        return desc

    def validate(self, value):
        if not isinstance(value, (list, tuple)):
            raise TypeError('Expected a list or tuple for "%s"' % self.name)

        for elementValue in value:
            self.element.validate(elementValue)

    def deserialize(self, value):
        if isinstance(value, str):
            value = value.split(self.separator)

        return [self.element.deserialize(v) for v in value]



from .choice import Choice


class StringMultichoice(Choice):
    """Define a multichose string parameter type.

    Values of this type are iterable sequences of strings
    all of which must be an element of a predefined set.

    >>> @argument('people', types.StringMultichoice, choices=('alice', 'bob', 'charlie'))
    ... def func(people=('alice', 'bob')):
    ...     pass
    """

    paramType = 'string-enumeration-multiple'
    multiple = True



import sys


def tee_stdout(cls):
    """Decorate a Tee class giving it access to sys.stdout.

    This function is intended to decorate a class that subclasses the
    Tee object. It assigns the _set_stream and _get_stream methods on
    the cls in such a way that the class's write() method will receive
    data passed to sys.stdout.
    """
    def _set_stdout(stream):
        """Set sys.stdout to a new file-like object.

        This is a private function for re-assigning sys.stdout. It returns
        the old value of sys.stdout.

        """
        old = sys.stdout
        sys.stdout = stream
        return old

    def _get_stdout():
        """Return the object that sys.stdout points too."""
        return sys.stdout

    cls._set_stream = staticmethod(_set_stdout)
    cls._get_stream = staticmethod(_get_stdout)

    return cls


def tee_stderr(cls):
    """Decorate a Tee class giving it access to sys.stderr.

    This function is intended to decorate a class that subclasses the
    Tee object. It assigns the _set_stream and _get_stream methods on
    the cls in such a way that the class's write() method will receive
    data passed to sys.stderr.
    """
    def _set_stderr(stream):
        """Set sys.stderr to a new file-like object.

        This is a private function for re-assigning sys.stderr. It returns
        the old value of sys.stderr.

        """
        old = sys.stderr
        sys.stderr = stream
        return old

    def _get_stderr():
        """Return the object that sys.stderr points too."""
        return sys.stderr

    cls._set_stream = staticmethod(_set_stderr)
    cls._get_stream = staticmethod(_get_stderr)
    return cls


class Tee:
    """Implements a context manager for intercepting write streams.

    This object is loosely inspired by the classic GNU utility
    tee(1). It allows a developer to intercept stream write() calls,
    act on that data, and then passes that data through to the
    original stream object.

    Consider the case of sys.stdout; which is like a global 'pointer'
    to a writeable object. It may be reassigned to other writable
    objects to 'hijack' the effects of print() or sys.stdout.write()
    statements. This is somewhat problematic, as we may want to write
    both to the new object, and to the previous value of sys.stdout
    (e.g. a pseudo-tty).

    The Tee object implements a basic, singly-linked list of writable
    objects. When a process calls print(), data is written to
    sys.stdout, which (if it is a Tee object) calls write(), and then
    passes the data to its downstream object for processing. This
    process continues until either it encounters a non-Tee object
    (e.g. the open stdout file object) or a Tee that does not pass
    data through to its downstream connections.

    """

    # Note that by abstracting over the stream via the _set_stream and
    # _get_stream methods we can use decorators like tee_stdout and
    # tee_stderr to implement common functionality in a base class and
    # then decorate derived subclasses which handle specific stream
    # functionality. See TeeStdOut and TeeStdErr.
    @staticmethod
    def _set_stream(stream):
        """Set a stream 'pointer' (e.g. sys.stdout) to a new writeable object.

        This function should set a known global pointer to a stream
        object (e.g sys.stdout) to a new writable object. it should
        return the old value of the pointer so that it may be stored
        in the '_downstream' attribute of the Tee object.

        """
        pass

    @staticmethod
    def _get_stream():
        """Get the stream 'pointer' this object is Teeing.

        For example, if this object was Teeing sys.stdout, it would
        return the current value of sys.stdout.
        """
        pass

    def __init__(self, pass_through=True):
        self.pass_through = pass_through
        self._downstream = self._set_stream(self)

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.reset()

    def reset(self):
        """Remove the current Tee object from the list of Tee objects.

        This removes the current object from the linked list of Tee
        objects associated with the stream returned by
        self._get_stream().

        Consider the case of several chained Tee objects (e.g. t1, t2,
        t3) where the original '<stdout>' open file is downstream from
        t1, t1 is downstream from t2, and t2 is downstream from t3:

        t3 => t2 => t1 => <open file '<stdout>', ...>
        ^
        sys.stdout

        sys.stdout points to t3. If we call t1.reset() it would be
        incorrect to set sys.stdout to the <open file ...> (t1's
        downstream). Instead sys.stdout must continue to point to t3
        and t2._downstream should be updated to point to the <open
        file ...>. This ensures that Tee objects may leave the chain
        as-needed without losing connections.

        """
        prev, cur = None, self._get_stream()

        if cur == self:
            self._set_stream(self._downstream)
            return

        while hasattr(cur, '_downstream'):
            prev, cur = cur, cur._downstream
            if cur == self:
                prev._downstream = cur._downstream

    def __getattr__(self, attr):
        return getattr(self._downstream, attr)

    def write(self, *args, **kwargs):
        if self.pass_through:
            self._downstream.write(*args, **kwargs)

    def flush(self, *args, **kwargs):
        if self.pass_through:
            self._downstream.flush(*args, **kwargs)


@tee_stdout
class TeeStdOut(Tee):
    pass


@tee_stderr
class TeeStdErr(Tee):
    pass



from .base import Base


class Boolean(Base):
    """Define a boolean task parameter.

    >>> @argument('debug', types.Boolean)
    ... def func(debug=False):
    ...     pass
    """

    description = {
        'type': 'boolean',
        'description': 'Provide a boolean value'
    }

    def serialize(self, value):
        return bool(value)



import os
import sys

import traceback as tb
from distutils.version import LooseVersion

from celery import Celery, __version__

from celery.signals import (
    before_task_publish,
    task_failure,
    task_postrun,
    task_prerun,
    task_revoked,
    task_success,
    worker_ready)


from girder_client import GirderClient

from girder_worker import logger
from girder_worker.context import get_context
from girder_worker.entrypoint import discover_tasks
from girder_worker.task import Task
from girder_worker.utils import (
    JobSpecNotFound,
    JobStatus,
    StateTransitionException,
    _job_manager,
    _update_status,
    is_builtin_celery_task,
    is_revoked
)
from girder_worker.utils.transform import ResultTransform

import jsonpickle
from kombu.serialization import register


@before_task_publish.connect  # noqa: C901
def girder_before_task_publish(sender=None, body=None, exchange=None,
                               routing_key=None, headers=None, properties=None,
                               declare=None, retry_policy=None, **kwargs):

    if is_builtin_celery_task(sender):
        return

    job = None

    try:
        context = get_context()
        if 'jobInfoSpec' not in headers and not headers.get('girder_job_disable'):
            job = context.create_task_job(
                Task.girder_job_defaults(), sender=sender, body=body, exchange=exchange,
                routing_key=routing_key, headers=headers, properties=properties, declare=declare,
                retry_policy=retry_policy, **kwargs)

        if 'girder_api_url' not in headers:
            context.attach_girder_api_url(sender=sender, body=body,
                                          exchange=exchange,
                                          routing_key=routing_key,
                                          headers=headers,
                                          properties=properties,
                                          declare=declare,
                                          retry_policy=retry_policy,
                                          **kwargs)

        if 'girder_client_token' not in headers:
            context.attach_girder_client_token(sender=sender,
                                               body=body,
                                               exchange=exchange,
                                               routing_key=routing_key,
                                               headers=headers,
                                               properties=properties,
                                               declare=declare,
                                               retry_policy=retry_policy,
                                               **kwargs)
        if 'girder_result_hooks' in headers:
            if job is not None:
                for result_hook in headers['girder_result_hooks']:
                    if isinstance(result_hook, ResultTransform):
                        result_hook.job = job

            # Celery task headers are not automatically serialized by celery
            # before being passed off to ampq for byte packing. We will have
            # to do that here.
            p = jsonpickle.pickler.Pickler()
            headers['girder_result_hooks'] = \
                [p.flatten(grh) for grh in headers['girder_result_hooks']]

        # Finally,  remove all reserved_options from headers
        for key in Task.reserved_options:
            headers.pop(key, None)
    except Exception:
        logger.exception('An error occurred in girder_before_task_publish.')
        raise


@worker_ready.connect
def check_celery_version(*args, **kwargs):
    if LooseVersion(__version__) < LooseVersion('4.0.0'):
        sys.exit("""You are running Celery {}.

girder-worker requires celery>=4.0.0""".format(__version__))


@task_prerun.connect
def gw_task_prerun(task=None, sender=None, task_id=None,
                   args=None, kwargs=None, **rest):
    """Deserialize the jobInfoSpec passed in through the headers.

    This provides the a JobManager class as an attribute of the
    task before task execution.  decorated functions may bind to
    their task and have access to the job_manager for logging and
    updating their status in girder.
    """
    if is_builtin_celery_task(sender.name):
        return

    try:
        task.job_manager = _job_manager(task.request, task.request.headers)
        _update_status(task, JobStatus.RUNNING)

    except JobSpecNotFound:
        task.job_manager = None
        logger.warning('No jobInfoSpec. Setting job_manager to None.')
    except StateTransitionException:
        # Fetch the current status of the job
        status = task.job_manager.refreshStatus()
        # If we are canceling we want to stay in that state
        if status != JobStatus.CANCELING:
            raise

    try:
        task.girder_client = GirderClient(apiUrl=task.request.girder_api_url)
        task.girder_client.token = task.request.girder_client_token
    except AttributeError:
        task.girder_client = None

    # Deserialize girder_result_hooks if they exist
    if hasattr(task.request, 'girder_result_hooks'):
        u = jsonpickle.unpickler.Unpickler()
        task.request.girder_result_hooks = \
            [u.restore(grh) for grh in task.request.girder_result_hooks]


@task_success.connect
def gw_task_success(sender=None, **rest):
    if is_builtin_celery_task(sender.name):
        return

    try:
        if is_revoked(sender):
            _update_status(sender, JobStatus.CANCELED)
        else:
            _update_status(sender, JobStatus.SUCCESS)
    except AttributeError:
        pass
    except StateTransitionException:
        # Fetch the current status of the job
        status = sender.job_manager.refreshStatus()
        # If we are in CANCELING move to CANCELED
        if status == JobStatus.CANCELING or is_revoked(sender):
            _update_status(sender, JobStatus.CANCELED)
        else:
            raise


@task_failure.connect
def gw_task_failure(sender=None, exception=None,
                    traceback=None, **rest):
    if is_builtin_celery_task(sender.name):
        return

    try:

        msg = '%s: %s\n%s' % (
            exception.__class__.__name__, exception,
            ''.join(tb.format_tb(traceback)))

        sender.job_manager.write(msg)
        _update_status(sender, JobStatus.ERROR)

    except AttributeError:
        pass


@task_postrun.connect
def gw_task_postrun(task=None, sender=None, task_id=None,
                    args=None, kwargs=None,
                    retval=None, state=None, **rest):
    try:
        task.job_manager._flush()
    except AttributeError:
        pass
    finally:
        # Release stdout/stderr
        if hasattr(task, 'job_manager') and \
           hasattr(task.job_manager, 'cleanup') and \
           callable(task.job_manager.cleanup):
            task.job_manager.cleanup()


@task_revoked.connect
def gw_task_revoked(sender=None, request=None, **rest):
    try:
        sender.job_manager = _job_manager(headers=request.message.headers,
                                          kwargs=request.kwargsrepr)
        _update_status(sender, JobStatus.CANCELED)
    except AttributeError:
        pass
    except JobSpecNotFound:
        logger.warning(
            "No jobInfoSpec. Unable to move \'%s\' into CANCELED state.")


register('girder_io', jsonpickle.encode, jsonpickle.decode,
         content_type='application/json',
         content_encoding='utf-8')

app = Celery(
    main=os.environ.get('GIRDER_WORKER_CELERY_APP_MAIN', 'app_main'),
    task_cls='girder_worker.task:Task')

discover_tasks(app)

app.config_from_object('girder_worker.celeryconfig', force=True)
